
from __future__ import print_function

import os
import sys
import _setup_test_env  # noqa

from returnn.datasets import Dataset
from returnn.datasets.hdf import *
from nose.tools import assert_equal
from nose.tools import assert_not_equal
from nose.tools import assert_raises
from nose.tools import raises
import returnn.util.basic as util
import h5py
import numpy as np
import os
import unittest
from returnn.util import better_exchook


my_dir = os.path.dirname(os.path.abspath(__file__))


class TestHDFDataset(object):
  @classmethod
  def setup_class(cls):
    """
    :return:
     This method is run once before starting testing
    """

  @classmethod
  def teardown_class(cls):
    """
    This method is run once after completing all tests
    :return:
    """

  def setup(self):
    """
    This method is run before each test is going to be started
    :return:
    """

  def teardown(self):
    """
    This method is run after finishing of each test
    :return:
    """

  def test_init(self):
    """
    This method tests initialization of the HDFDataset class
    """
    toy_dataset = HDFDataset()
    assert_equal(toy_dataset.file_start, [0], "self.file_start init problem, should be [0]")
    assert_equal(toy_dataset.files, [], "self.files init problem, should be []")
    assert_equal(toy_dataset.file_seq_start, [], "self.file_seq_start init problem, should be []")
    return toy_dataset

  def test_addfile(self):
    """
    This method tests self.addfile function
    """
    toy_dataset = self.test_init()
    # TODO: auto-generate file, then use here
    #toy_dataset.add_file("/u/kulikov/develop/crnn/tests/toy_set.hdf")


def generate_dummy_hdf(num_datasets=1):
  filenames = []
  for idx in range(1, num_datasets + 1):
    fn = get_test_tmp_file(".%i.hdf5" % idx)
    filenames.append(fn)
    dataset = h5py.File(fn, 'w')
    dataset.create_group('streams')

    dataset['streams'].create_group('features')
    dataset['streams']['features'].attrs['parser'] = "feature_sequence"
    dataset['streams']['features'].create_group('data')

    dataset['streams'].create_group('classes')
    dataset['streams']['classes'].attrs['parser'] = "sparse"
    dataset['streams']['classes'].create_group('data')

    import string
    random = np.random.RandomState()
    feature_size = 13
    seq_names = ['dataset_%d_sequence_%d' % (idx, i) for i in range(100)]
    class_names = list(string.ascii_lowercase)
    num_classes = len(class_names)
    print(class_names, len(class_names))
    for idx in range(100):
      class_id = random.randint(low=0, high=num_classes)
      seq_len  = random.randint(low=1, high=20)
      features = random.rand(seq_len, feature_size)
      dataset['streams']['features']['data'].create_dataset(name=seq_names[idx], data=features, dtype='float32')
      dataset['streams']['classes']['data'].create_dataset(name=seq_names[idx], shape=(1,), data=np.int32(class_id), dtype='int32')

    dt = h5py.special_dtype(vlen=str)
    feature_names = dataset['streams']['classes'].create_dataset("feature_names", shape=(len(class_names),), dtype=dt)
    for id_x, orth in enumerate(class_names):
      feature_names[id_x] = orth

    dt = h5py.special_dtype(vlen=str)
    sequence_names_data = dataset.create_dataset("seq_names", shape=(len(seq_names),), dtype=dt)
    for ind, val in enumerate(seq_names):
      sequence_names_data[ind] = val

    dataset.close()
  return filenames


# Note that nosetests might even call this function, as it has "test" in its name... Does not matter, though.
def get_test_tmp_file(suffix=".hdf"):
  """
  :param str suffix: e.g. ".hdf"
  :return: filename
  :rtype: str
  """
  import tempfile
  f = tempfile.NamedTemporaryFile(suffix=suffix, delete=False)
  f.close()
  fn = f.name
  import atexit
  atexit.register(lambda: os.remove(fn))
  return fn


_hdf_cache = {}  # opts -> hdf fn


def generate_hdf_from_other(opts, suffix=".hdf"):
  """
  :param dict[str] opts:
  :param str suffix:
  :return: hdf filename
  :rtype: str
  """
  # See test_hdf_dump.py and tools/hdf_dump.py.
  from returnn.util.basic import make_hashable
  cache_key = make_hashable(opts)
  if cache_key in _hdf_cache:
    return _hdf_cache[cache_key]
  fn = get_test_tmp_file(suffix=suffix)
  from returnn.datasets.basic import init_dataset
  dataset = init_dataset(opts)
  hdf_dataset = HDFDatasetWriter(fn)
  hdf_dataset.dump_from_dataset(dataset)
  hdf_dataset.close()
  _hdf_cache[cache_key] = fn
  return fn


def generate_hdf_from_dummy():
  """
  :return: hdf filename
  :rtype: str
  """
  return generate_hdf_from_other(
    {"class": "DummyDataset", "input_dim": 13, "output_dim": 7, "num_seqs": 23, "seq_len": 17})


def test_hdf_dump():
  generate_hdf_from_dummy()


class DatasetTestReader:
  def __init__(self, dataset):
    """
    :param Dataset dataset:
    """
    self.dataset = dataset
    self.data_keys = None
    self.data_shape = {}  # key -> shape
    self.data_sparse = {}  # key -> bool
    self.data_dtype = {}  # key -> str
    self.data = {}  # type: typing.Dict[str,typing.List[numpy.ndarray]]  # key -> list
    self.seq_lens = []  # type: typing.List[Util.NumbersDict]
    self.seq_tags = []
    self.num_seqs = 0

  def read_all(self, epoch=1):
    """
    :param int epoch:
    """
    dataset = self.dataset
    dataset.init_seq_order(epoch=epoch)
    data_keys = dataset.get_data_keys()
    self.data_keys = data_keys
    self.data_shape = {key: dataset.get_data_shape(key) for key in data_keys}
    self.data_sparse = {key: dataset.is_data_sparse(key) for key in data_keys}
    self.data_dtype = {key: dataset.get_data_dtype(key) for key in data_keys}
    seq_idx = 0
    while dataset.is_less_than_num_seqs(seq_idx):
      dataset.load_seqs(seq_idx, seq_idx + 1)
      for key in data_keys:
        data = dataset.get_data(seq_idx=seq_idx, key=key)
        self.data.setdefault(key, []).append(data)
      seq_len = dataset.get_seq_length(seq_idx=seq_idx)
      self.seq_lens.append(seq_len)
      seq_tag = dataset.get_tag(seq_idx)
      self.seq_tags.append(seq_tag)
      seq_idx += 1
    print("Iterated through %r, num seqs %i" % (dataset, seq_idx))
    self.num_seqs += seq_idx


def test_hdf_dump_not_frame_synced():
  num_seqs = 3
  from returnn.datasets.generating import TaskNumberBaseConvertDataset
  hdf_fn = generate_hdf_from_other({"class": "TaskNumberBaseConvertDataset", "num_seqs": num_seqs})
  hdf = HDFDataset([hdf_fn])
  orig = TaskNumberBaseConvertDataset(num_seqs=num_seqs)
  hdf_reader = DatasetTestReader(hdf)
  orig_reader = DatasetTestReader(orig)
  hdf_reader.read_all()
  orig_reader.read_all()
  assert hdf_reader.data_keys == orig_reader.data_keys == ["data", "classes"]
  assert hdf_reader.num_seqs == orig_reader.num_seqs == num_seqs
  for seq_idx in range(num_seqs):
    # Not synced, i.e. different lengths:
    assert_not_equal(orig_reader.seq_lens[seq_idx]["data"], orig_reader.seq_lens[seq_idx]["classes"])
    for key in orig_reader.data_keys:
      assert_equal(hdf_reader.seq_lens[seq_idx][key], orig_reader.seq_lens[seq_idx][key])
      assert_equal(hdf_reader.data[key][seq_idx].tolist(), orig_reader.data[key][seq_idx].tolist())


def test_HDFDataset_partition_epoch():
  partition_epoch = 3
  num_seqs = 11
  from returnn.datasets.generating import TaskNumberBaseConvertDataset
  hdf_fn = generate_hdf_from_other({"class": "TaskNumberBaseConvertDataset", "num_seqs": num_seqs})
  hdf = HDFDataset([hdf_fn], partition_epoch=partition_epoch)
  orig = TaskNumberBaseConvertDataset(num_seqs=num_seqs)
  hdf_reader = DatasetTestReader(hdf)
  orig_reader = DatasetTestReader(orig)
  for epoch in range(1, partition_epoch + 1):
    hdf_reader.read_all(epoch=epoch)
  orig_reader.read_all()  # single epoch
  assert hdf_reader.data_keys == orig_reader.data_keys == ["data", "classes"]
  assert hdf_reader.num_seqs == orig_reader.num_seqs == num_seqs
  for seq_idx in range(num_seqs):
    # Not synced, i.e. different lengths:
    assert_not_equal(orig_reader.seq_lens[seq_idx]["data"], orig_reader.seq_lens[seq_idx]["classes"])
    for key in orig_reader.data_keys:
      assert_equal(hdf_reader.seq_lens[seq_idx][key], orig_reader.seq_lens[seq_idx][key])
      assert_equal(hdf_reader.data[key][seq_idx].tolist(), orig_reader.data[key][seq_idx].tolist())


def test_SimpleHDFWriter():
  fn = get_test_tmp_file(suffix=".hdf")
  os.remove(fn)  # SimpleHDFWriter expects that the file does not exist
  n_dim = 13
  writer = SimpleHDFWriter(filename=fn, dim=n_dim, labels=None)
  seq_lens1 = [11, 7, 5]
  writer.insert_batch(
    inputs=numpy.random.normal(size=(len(seq_lens1), max(seq_lens1), n_dim)).astype("float32"),
    seq_len=seq_lens1,
    seq_tag=["seq-%i" % i for i in range(len(seq_lens1))])
  seq_lens2 = [10, 13, 3, 2]
  writer.insert_batch(
    inputs=numpy.random.normal(size=(len(seq_lens2), max(seq_lens2), n_dim)).astype("float32"),
    seq_len=seq_lens2,
    seq_tag=["seq-%i" % (i + len(seq_lens1)) for i in range(len(seq_lens2))])
  writer.close()
  seq_lens = seq_lens1 + seq_lens2

  dataset = HDFDataset(files=[fn])
  reader = DatasetTestReader(dataset=dataset)
  reader.read_all()
  assert "data" in reader.data_keys  # "classes" might be in there as well, although not really correct/existing
  assert reader.data_sparse["data"] is False
  assert list(reader.data_shape["data"]) == [n_dim]
  assert reader.data_dtype["data"] == "float32"
  assert len(seq_lens) == reader.num_seqs
  for i, seq_len in enumerate(seq_lens):
    assert reader.seq_lens[i]["data"] == seq_len
  print("tags:", reader.seq_tags)
  assert_equal(reader.seq_tags, ["seq-%i" % i for i in range(reader.num_seqs)])
  assert isinstance(reader.seq_tags[0], str)


def test_SimpleHDFWriter_small():
  fn = get_test_tmp_file(suffix=".hdf")
  os.remove(fn)  # SimpleHDFWriter expects that the file does not exist
  n_dim = 3
  writer = SimpleHDFWriter(filename=fn, dim=n_dim, labels=None)
  seq_lens = [2, 3]
  writer.insert_batch(
    inputs=numpy.random.normal(size=(len(seq_lens), max(seq_lens), n_dim)).astype("float32"),
    seq_len=seq_lens,
    seq_tag=["seq-%i" % i for i in range(len(seq_lens))])
  writer.close()

  dataset = HDFDataset(files=[fn])
  reader = DatasetTestReader(dataset=dataset)
  reader.read_all()
  assert "data" in reader.data_keys  # "classes" might be in there as well, although not really correct/existing
  assert reader.data_sparse["data"] is False
  assert list(reader.data_shape["data"]) == [n_dim]
  assert reader.data_dtype["data"] == "float32"
  assert len(seq_lens) == reader.num_seqs
  for i, seq_len in enumerate(seq_lens):
    assert reader.seq_lens[i]["data"] == seq_len

  if sys.version_info[0] >= 3:  # gzip.compress is >=PY3
    print("raw content (gzipped):")
    import gzip
    print(repr(gzip.compress(open(fn, "rb").read())))


def test_read_simple_hdf():
  if sys.version_info[0] <= 2:  # gzip.decompress is >=PY3
    raise unittest.SkipTest
  # n_dim, seq_lens, raw_gzipped is via test_SimpleHDFWriter_small
  n_dim = 3
  seq_lens = [2, 3]
  raw_gzipped = (
    b'\x1f\x8b\x08\x00\x80\xc8f\\\x02\xff\xed\x9a=l\xd3@\x14\xc7\xefl\'XQ\x8a\xd2vhTT\xf0\x84X\x90\xc2V1\xb8TJ\xa0C'
    b'\x81\x8a\x04\xa9\x03BM\x85i"5Q\x82]$`h%\x18@\xea\xc6\x02\x1b\x03\x03\x1f\x0b\x0b\x1bm\x10S\xa7NLL\xb01\x96\x1d'
    b'\xa9\xd8\xbewI}\x8d\x9d\xb1i\xf9\xff\x86\\\xee\xf2\x9e\xef\xc3\xff\xbb{'
    b'\xf6\xe5\xc5\\\xf1\xeaHf2\xc3\x02L\x93\x19,'
    b'\xc7\x0e\xb2O|\xb7\xa3y\xf9\xfb\x12\xa5\x9c\xd2\xe7\x94\xbe\xd3d\xb9\x19\xfe\x96\xa7\xf2\x1c]\xdf\xd2E\xbeF\x8e'
    b'\x95[\xa5R`\xbd\xaf \xeby\x95\x12\xe9\x05\x06\xfeG\xe6J\xb3\x0bA\xbaH\xf9\x02\xa5;Z\xd4n\xb5\xba\xec\xac\xba'
    b'\x8c\xb9N{\xdei\xaex5W\x94\xd7\x9b\xad5O\x94W\xaa+nW\xaf\x83\xf44JzUu\x9de\xd3\xfe\\\t\x14{'
    b'\xda\xffn\x8a\xeb/T=\xaf\\\x7f\xec\x04:7\xfd\xe9\x14Z^\x89\xcc\x0f\x92\xbd\xefS '
    b'\x7f3\xf4o\xae5\x8a\xf5\x86\x1b\xeb\xc7\x99ZoF\xfa\xcdS\x97\xc5\xfc\x1aX\xaf\xf4\x1f\x91\xfe\x95z\xc3q=\xa7\xe5'
    b'&\xf9\xa7\xe2\xdb]v\xda\xf1\xed\xee\xdd\x9e|\xe28s\x96\x16>\\\xe6\x85=\xe7\xbc\xaf\xbdN\xeb\xca8\x17mKQ^\xd3'
    b'\xb4\xd0\xc1$\x7f\x9d\xab+\x89`\x8cZ\x1b\x18o\xec\xdc\xbf\xd3\xbb\xc3\xc3A\xf9\xc6\xcd"\xf7G\xda\x92:\x1dK\xb67'
    b'\xe5\xfak$\xdb\xc9\xd5\xfdg:\xd9N\xce\x8b\xd6\xd4\xf1^7\x0e\xebJ\xf4\x8c\x0b\x99t\xf5\xa9)\xfb\x9d\xd6\xd5Y.t\r'
    b'\xf4-\x86\xd6\xa2\xf9@z\xd3YTo\x9a\xbew\x8a|\rqY\xa3\xbf\xdeZG2\x1e\xc1>\xcb\xfb\xed\xb3V\xb2\xdf\xe6Y\xa5\xc0'
    b'\x88\x8e\x9b\x81-\n\x00\x00\x00\x00\xe0D1('
    b'\x8eN)\xcf\x99j|\xa9\xfb\xf1q`9j\x9d\xeb\xc6\xd1\x13&\x9bX\xef>_\xc6\xc6\xd3\xd3\xe3\xbdPS\x8f\x8f\xa7kG\x10G'
    b'\xeb\x87\xfa\x99\x1f\xe0\xb7iG\x9f\x86\xb5\x18\xbb\xb7\x8a]\\|\xfd\xc5\x8e\xe6\xd3\xca}@\\\x0e\x00\x00\x00'
    b'\x000lqt\xf4\x9cC}\x1f}\xf0\x9c#\xcd\x92\xce96\xe8\rm6R\xdf\xb0\x9fs\xb8N\xfbb!\xfc\xbc\x14\xf6Y\x06\xf9:\xa4'
    b'\x02\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x1cK>\xde>\xdf\xb9\xf6\xc3\xec\xfc'
    b'}\xf0\xc7\xbe7\xf5m\xe6\xf2\xe7\xa7\xf6^\xee\xc9\xd6\xefG\xb3\x9d\x87\xcf\x9c\xed_\x1f&\xbf\xbe\xd4\xcet\xae'
    b'\xaf\xbf\x99\xc9~\xda\xdd\xaa/\xbf\xb7\xef.\xeen\xbf^2"\xff\x82\xfd\x07\xf7a\x8c\xa2\xd4>\x00\x00')
  import gzip
  fn = get_test_tmp_file(suffix=".hdf")
  with open(fn, "wb") as f:
    f.write(gzip.decompress(raw_gzipped))

  dataset = HDFDataset(files=[fn])
  reader = DatasetTestReader(dataset=dataset)
  reader.read_all()
  print("tags:", reader.seq_tags)
  assert len(seq_lens) == reader.num_seqs
  assert_equal(reader.seq_tags, ["seq-0", "seq-1"])
  for i, seq_len in enumerate(seq_lens):
    assert reader.seq_lens[i]["data"] == seq_len
  assert "data" in reader.data_keys  # "classes" might be in there as well, although not really correct/existing
  assert reader.data_sparse["data"] is False
  assert list(reader.data_shape["data"]) == [n_dim]
  assert reader.data_dtype["data"] == "float32"


def test_SimpleHDFWriter_ndim1_var_len():
  # E.g. attention weights, shape (dec-time,enc-time) per seq.
  fn = get_test_tmp_file(suffix=".hdf")
  os.remove(fn)  # SimpleHDFWriter expects that the file does not exist
  writer = SimpleHDFWriter(filename=fn, dim=None, ndim=2, labels=None)
  dec_seq_lens1 = [11, 7, 5]
  enc_seq_lens1 = [13, 6, 8]
  batch1_data = numpy.random.normal(size=(len(dec_seq_lens1), max(dec_seq_lens1), max(enc_seq_lens1))).astype("float32")
  writer.insert_batch(
    inputs=batch1_data,
    seq_len={0: dec_seq_lens1, 1: enc_seq_lens1},
    seq_tag=["seq-%i" % i for i in range(len(dec_seq_lens1))])
  dec_seq_lens2 = [10, 13, 3, 2]
  enc_seq_lens2 = [11, 13, 5, 4]
  batch2_data = numpy.random.normal(size=(len(dec_seq_lens2), max(dec_seq_lens2), max(enc_seq_lens2))).astype("float32")
  writer.insert_batch(
    inputs=batch2_data,
    seq_len={0: dec_seq_lens2, 1: enc_seq_lens2},
    seq_tag=["seq-%i" % (i + len(dec_seq_lens1)) for i in range(len(dec_seq_lens2))])
  writer.close()
  dec_seq_lens = dec_seq_lens1 + dec_seq_lens2
  enc_seq_lens = enc_seq_lens1 + enc_seq_lens2

  dataset = HDFDataset(files=[fn])
  reader = DatasetTestReader(dataset=dataset)
  reader.read_all()
  assert "data" in reader.data_keys  # "classes" might be in there as well, although not really correct/existing
  assert reader.data_sparse["data"] is False
  assert list(reader.data_shape["data"]) == []
  assert reader.data_dtype["data"] == "float32"
  assert "sizes" in reader.data_keys
  # sizes sparsity does not really matter...
  assert reader.data_dtype["sizes"] == "int32"
  assert list(reader.data_shape["sizes"]) == []
  assert len(dec_seq_lens) == len(enc_seq_lens) == reader.num_seqs
  for i, (dec_seq_len, enc_seq_len) in enumerate(zip(dec_seq_lens, enc_seq_lens)):
    assert reader.seq_lens[i]["data"] == dec_seq_len * enc_seq_len
    assert reader.seq_lens[i]["sizes"] == 2
    assert reader.data["sizes"][i].tolist() == [dec_seq_len, enc_seq_len], "got %r" % (
      reader.data["sizes"][i],)


def test_SimpleHDFWriter_extend_existing_file():
  fn = get_test_tmp_file(suffix=".hdf")
  os.remove(fn)  # SimpleHDFWriter expects that the file does not exist
  n_dim = 3
  writer = SimpleHDFWriter(filename=fn, dim=n_dim, labels=None)
  seq_lens = [2, 3]
  writer.insert_batch(
    inputs=numpy.random.normal(size=(len(seq_lens), max(seq_lens), n_dim)).astype("float32"),
    seq_len=seq_lens,
    seq_tag=["seq-%i" % i for i in range(len(seq_lens))])
  writer.close()
  assert os.path.exists(fn)

  writer = SimpleHDFWriter(filename=fn, dim=n_dim, labels=None, extend_existing_file=True)
  seq_lens2 = [4, 3, 2]
  writer.insert_batch(
    inputs=numpy.random.normal(size=(len(seq_lens2), max(seq_lens2), n_dim)).astype("float32"),
    seq_len=seq_lens2,
    seq_tag=["seq-%i" % (i + len(seq_lens)) for i in range(len(seq_lens2))])
  writer.close()
  seq_lens += seq_lens2

  dataset = HDFDataset(files=[fn])
  reader = DatasetTestReader(dataset=dataset)
  reader.read_all()
  assert "data" in reader.data_keys  # "classes" might be in there as well, although not really correct/existing
  assert reader.data_sparse["data"] is False
  assert list(reader.data_shape["data"]) == [n_dim]
  assert reader.data_dtype["data"] == "float32"
  assert len(seq_lens) == reader.num_seqs
  for i, seq_len in enumerate(seq_lens):
    assert reader.seq_lens[i]["data"] == seq_len


@unittest.skip("unfinished...")
def test_SimpleHDFWriter_swmr():
  fn = get_test_tmp_file(suffix=".hdf")
  os.remove(fn)  # SimpleHDFWriter expects that the file does not exist
  rnd = numpy.random.RandomState(42)
  n_dim = 13
  writer = SimpleHDFWriter(filename=fn, dim=n_dim, labels=None, swmr=True)

  # TODO
  # As we directly want to read it, we must know all the seqs (and seq tags) in advance.
  # Not only that; we also would need all seq lens in advance, to have the offsets in the HDF.
  # (Or do we? We could extend MetaDataset to ignore missing.
  #  But then we would need to enforce a reload at the next epoch. Which we can do.)
  # In any case though, this would not support overwriting existing seqs,
  # but we need that...
  # How to solve this? Different HDF format? But not really possible even in principle in a nice way, or is it?
  # Or write to new HDF file for each epoch?
  # Once finished through whole dataset (all sub epochs), delete old and replace?
  # Then we also don't really need swmr.

  reader = HDFDataset(files=[fn])  # TODO use swmr?
  reader.init_seq_order(epoch=1)

  seq_lens = [11, 7, 5, 10, 13, 3, 2]
  # Construct seq tags in a way that the lengths vary.
  seq_tags = [
    ("seq-%i-" % i) + "".join(["abcdefghijk"[rnd.randint(0, 10)] for _ in range(rnd.randint(1, 50))])
    for i in range(len(seq_lens))]
  seqs = numpy.random.normal(size=(len(seq_lens), max(seq_lens), n_dim)).astype("float32")
  for (s, e) in [(0, 3), (3, len(seq_lens))]:
    writer.insert_batch(
      inputs=seqs[s:e, :max(seq_lens[s:e])],
      seq_len=seq_lens[s:e],
      seq_tag=seq_tags[s:e])
    writer.flush()  # TODO when do we want it?

    # Now read.
    for seq_idx in range(s, e):
      reader.load_seqs(seq_idx, seq_idx + 1)
      # TODO read and check whether it is correct

  writer.close()  # Should not matter.


def dummy_iter_dataset(dataset):
  """
  :param Dataset dataset:
  """
  dataset.init_seq_order(epoch=1)
  data_keys = dataset.get_data_keys()
  seq_idx = 0
  while dataset.is_less_than_num_seqs(seq_idx):
    dataset.load_seqs(seq_idx, seq_idx + 1)
    for key in data_keys:
      dataset.get_data(seq_idx=seq_idx, key=key)
      dataset.get_tag(seq_idx)
    seq_idx += 1
  print("Iterated through %r, num seqs %i" % (dataset, seq_idx))


def test_hdf_simple_iter():
  hdf_fn = generate_hdf_from_dummy()
  dataset = HDFDataset(files=[hdf_fn])
  dataset.initialize()
  dummy_iter_dataset(dataset)


def test_hdf_simple_iter_cached():
  hdf_fn = generate_hdf_from_dummy()
  dataset = HDFDataset(files=[hdf_fn], cache_byte_size=100)
  dataset.initialize()
  dummy_iter_dataset(dataset)


def test_rnn_getCacheByteSizes_zero():
  from returnn.config import Config
  config = Config({"cache_size": "0"})
  import returnn.__main__ as rnn
  rnn.config = config
  sizes = rnn.get_cache_byte_sizes()
  assert len(sizes) == 3
  assert all([s == 0 for s in sizes])


def test_rnn_initData():
  hdf_fn = generate_hdf_from_dummy()
  from returnn.config import Config
  config = Config({"cache_size": "0", "train": hdf_fn, "dev": hdf_fn})
  import returnn.__main__ as rnn
  rnn.config = config
  rnn.init_data()
  train, dev = rnn.train_data, rnn.dev_data
  assert train and dev
  assert isinstance(train, HDFDataset)
  assert isinstance(dev, HDFDataset)
  assert train.cache_byte_size_total_limit == dev.cache_byte_size_total_limit == 0
  assert train.cache_byte_size_limit_at_start == dev.cache_byte_size_limit_at_start == 0


def test_hdf_no_cache_iter():
  hdf_fn = generate_hdf_from_dummy()
  dataset = HDFDataset(files=[hdf_fn])
  dataset.initialize()
  assert dataset.cache_byte_size_limit_at_start == 0
  assert dataset.cache_byte_size_total_limit == 0

  class DummyCallback:
    def __init__(self):
      self.was_called = False

    def __call__(self, *args, **kwargs):
      print("DummyCallback called!")
      self.was_called = True

  assert hasattr(dataset, "_preload_seqs")
  dataset._preload_seqs = DummyCallback()

  dummy_iter_dataset(dataset)
  import time
  time.sleep(1)  # maybe threads are running in background...
  assert not dataset._preload_seqs.was_called


def test_hdf_data_short_int_dtype():
  from returnn.datasets.generating import StaticDataset
  dataset = StaticDataset([
    {"data": numpy.array([1, 2, 3], dtype="uint8"), "classes": numpy.array([-1, 5], dtype="int16")}],
    output_dim={"data": (255, 1), "classes": (10, 1)})
  orig_data_dtype = dataset.get_data_dtype("data")
  orig_classes_dtype = dataset.get_data_dtype("classes")
  assert orig_data_dtype == "uint8" and orig_classes_dtype == "int16"

  hdf_fn = get_test_tmp_file(suffix=".hdf")
  hdf_writer = HDFDatasetWriter(filename=hdf_fn)
  hdf_writer.dump_from_dataset(dataset, use_progress_bar=False)
  hdf_writer.close()

  hdf_dataset = HDFDataset(files=[hdf_fn])
  hdf_dataset.initialize()
  hdf_dataset.init_seq_order(epoch=1)
  hdf_data_dtype = hdf_dataset.get_data_dtype("data")
  hdf_classes_dtype = hdf_dataset.get_data_dtype("classes")
  assert hdf_data_dtype == orig_data_dtype and hdf_classes_dtype == orig_classes_dtype
  hdf_data_dim = hdf_dataset.get_data_dim("data")
  hdf_classes_dim = hdf_dataset.get_data_dim("classes")
  assert hdf_data_dim == 255 and hdf_classes_dim == 10
  hdf_data_shape = hdf_dataset.get_data_shape("data")
  hdf_classes_shape = hdf_dataset.get_data_shape("classes")
  assert hdf_data_shape == [] and hdf_classes_shape == []
  hdf_dataset.load_seqs(0, 1)
  hdf_data_data = hdf_dataset.get_data(0, "data")
  hdf_data_classes = hdf_dataset.get_data(0, "classes")
  assert hdf_data_data.dtype == orig_data_dtype and hdf_data_classes.dtype == orig_classes_dtype


def test_hdf_data_target_int32():
  from returnn.datasets.generating import StaticDataset
  dataset = StaticDataset([
    {"data": numpy.array([1, 2, 3], dtype="uint8"),
     "classes": numpy.array([2147483647, 2147483646, 2147483645], dtype="int32")}],
    output_dim={"data": (255, 1), "classes": (10, 1)})
  dataset.initialize()
  dataset.init_seq_order(epoch=0)
  dataset.load_seqs(0, 1)
  orig_classes_dtype = dataset.get_data_dtype("classes")
  orig_classes_seq = dataset.get_data(0, "classes")
  assert orig_classes_seq.shape == (3,) and orig_classes_seq[0] == 2147483647
  assert orig_classes_seq.dtype == orig_classes_dtype == "int32"

  hdf_fn = get_test_tmp_file(suffix=".hdf")
  hdf_writer = HDFDatasetWriter(filename=hdf_fn)
  hdf_writer.dump_from_dataset(dataset, use_progress_bar=False)
  hdf_writer.close()

  hdf_dataset = HDFDataset(files=[hdf_fn])
  hdf_dataset.initialize()
  hdf_dataset.init_seq_order(epoch=1)
  hdf_classes_dtype = hdf_dataset.get_data_dtype("classes")
  assert hdf_classes_dtype == orig_classes_dtype
  hdf_classes_shape = hdf_dataset.get_data_shape("classes")
  assert hdf_classes_shape == []
  hdf_dataset.load_seqs(0, 1)
  hdf_data_classes = hdf_dataset.get_data(0, "classes")
  assert hdf_data_classes.dtype == orig_classes_dtype
  assert all(hdf_data_classes == orig_classes_seq)


def test_hdf_target_float_dtype():
  from returnn.datasets.generating import StaticDataset
  dataset = StaticDataset([
    {"data": numpy.array([1, 2, 3], dtype="float32"), "classes": numpy.array([-1, 5], dtype="float32")}],
    output_dim={"data": (1, 1), "classes": (1, 1)})
  orig_data_dtype = dataset.get_data_dtype("data")
  orig_classes_dtype = dataset.get_data_dtype("classes")
  assert orig_data_dtype == "float32" and orig_classes_dtype == "float32"

  hdf_fn = get_test_tmp_file(suffix=".hdf")
  hdf_writer = HDFDatasetWriter(filename=hdf_fn)
  hdf_writer.dump_from_dataset(dataset, use_progress_bar=False)
  hdf_writer.close()

  hdf_dataset = HDFDataset(files=[hdf_fn])
  hdf_dataset.initialize()
  hdf_dataset.init_seq_order(epoch=1)
  hdf_data_dtype = hdf_dataset.get_data_dtype("data")
  hdf_classes_dtype = hdf_dataset.get_data_dtype("classes")
  assert hdf_data_dtype == orig_data_dtype and hdf_classes_dtype == orig_classes_dtype
  hdf_data_dim = hdf_dataset.get_data_dim("data")
  hdf_classes_dim = hdf_dataset.get_data_dim("classes")
  assert hdf_data_dim == 1 and hdf_classes_dim == 1
  hdf_data_shape = hdf_dataset.get_data_shape("data")
  hdf_classes_shape = hdf_dataset.get_data_shape("classes")
  assert hdf_data_shape == [] and hdf_classes_shape == []
  hdf_dataset.load_seqs(0, 1)
  hdf_data_data = hdf_dataset.get_data(0, "data")
  hdf_data_classes = hdf_dataset.get_data(0, "classes")
  assert hdf_data_data.dtype == orig_data_dtype and hdf_data_classes.dtype == orig_classes_dtype


def test_hdf_target_float_dense():
  from returnn.datasets.generating import StaticDataset
  dataset = StaticDataset([
    {"data": numpy.array([[1, 2, 3], [2, 3, 4]], dtype="float32"),
     "classes": numpy.array([[-1, 5], [-2, 4], [-3, 2]], dtype="float32")}])
  orig_data_dtype = dataset.get_data_dtype("data")
  orig_classes_dtype = dataset.get_data_dtype("classes")
  assert orig_data_dtype == "float32" and orig_classes_dtype == "float32"
  orig_data_shape = dataset.get_data_shape("data")
  orig_classes_shape = dataset.get_data_shape("classes")
  assert orig_data_shape == [3] and orig_classes_shape == [2]

  hdf_fn = get_test_tmp_file(suffix=".hdf")
  hdf_writer = HDFDatasetWriter(filename=hdf_fn)
  hdf_writer.dump_from_dataset(dataset, use_progress_bar=False)
  hdf_writer.close()

  hdf_dataset = HDFDataset(files=[hdf_fn])
  hdf_dataset.initialize()
  hdf_dataset.init_seq_order(epoch=1)
  hdf_data_dtype = hdf_dataset.get_data_dtype("data")
  hdf_classes_dtype = hdf_dataset.get_data_dtype("classes")
  assert hdf_data_dtype == orig_data_dtype and hdf_classes_dtype == orig_classes_dtype
  hdf_data_dim = hdf_dataset.get_data_dim("data")
  hdf_classes_dim = hdf_dataset.get_data_dim("classes")
  assert hdf_data_dim == orig_data_shape[-1] and hdf_classes_dim == orig_classes_shape[-1]
  hdf_data_shape = hdf_dataset.get_data_shape("data")
  hdf_classes_shape = hdf_dataset.get_data_shape("classes")
  assert hdf_data_shape == orig_data_shape and hdf_classes_shape == orig_classes_shape
  hdf_dataset.load_seqs(0, 1)
  hdf_data_data = hdf_dataset.get_data(0, "data")
  hdf_data_classes = hdf_dataset.get_data(0, "classes")
  assert hdf_data_data.dtype == orig_data_dtype and hdf_data_classes.dtype == orig_classes_dtype


def test_HDFDataset_no_cache_efficiency():
  hdf_fn = generate_hdf_from_other({"class": "Task12AXDataset", "num_seqs": 23})
  hdf_dataset = HDFDataset(files=[hdf_fn], cache_byte_size=0)
  hdf_dataset.initialize()
  hdf_dataset.init_seq_order(epoch=1)
  hdf_dataset.load_seqs(0, 1)
  hdf_dataset.load_seqs(0, 2)
  hdf_dataset.load_seqs(0, 5)
  hdf_dataset.load_seqs(1, 2)
  hdf_dataset.load_seqs(1, 3)
  hdf_dataset.load_seqs(1, 7)
  hdf_dataset.load_seqs(3, 7)
  hdf_dataset.load_seqs(4, 10)
  # TODO... check alloc intervals etc


def test_siamese_triplet_sampling():
  datasets_path = generate_dummy_hdf(3)
  dataset = SiameseHDFDataset(input_stream_name="features", seq_label_stream="classes", files=datasets_path)

  dataset.initialize()
  for iter in range(1, 31):
    print("Initializing triplets... iteration %d" % iter)
    dataset.init_seq_order(epoch=iter)

    triplets = dataset.curr_epoch_triplets
    anchor_seq_names = [dataset.all_seq_names[id[0]] for id in triplets]
    same_class_seq_names = [dataset.all_seq_names[id[1]] for id in triplets]
    diff_class_seq_names = [dataset.all_seq_names[id[2]] for id in triplets]

    anchor_class = [dataset.seq_to_target[seq_id] for seq_id in anchor_seq_names]
    same_class = [dataset.seq_to_target[seq_id] for seq_id in same_class_seq_names]
    diff_class = [dataset.seq_to_target[seq_id] for seq_id in diff_class_seq_names]

    print("Testing pair sequences to belong to the same class...")
    assert (all(ac == same_class[id] for id, ac in enumerate(anchor_class)))
    print("Testing third element in a triplet to belong to a different class...")
    assert (all(ac != diff_class[id] for id, ac in enumerate(anchor_class)))
    print("------------------------------------------------------")

  print("Deleting temporary files...")
  for path in datasets_path:
    os.remove(path)
  print("Done.")


def test_siamese_collect_single_seq():
  datasets_path = generate_dummy_hdf(3)
  dataset = SiameseHDFDataset(input_stream_name="features", seq_label_stream="classes", files=datasets_path)

  dataset.initialize()
  dataset.init_seq_order(epoch=1)

  random = np.random.RandomState()
  seq_idx = random.randint(low=0, high=len(dataset.seq_name_to_idx))
  dataset_seq = dataset._collect_single_seq(seq_idx)
  print("Verify that single sequence consists of a triplet...")
  print(dataset_seq.features.keys())

  print("Deleting temporary files...")
  for path in datasets_path:
    os.remove(path)
  print("Done.")


if __name__ == "__main__":
  better_exchook.install()
  if len(sys.argv) <= 1:
    for k, v in sorted(globals().items()):
      if k.startswith("test_"):
        print("-" * 40)
        print("Executing: %s" % k)
        try:
          v()
        except unittest.SkipTest as exc:
          print("SkipTest:", exc)
        print("-" * 40)
    print("Finished all tests.")
  else:
    assert len(sys.argv) >= 2
    for arg in sys.argv[1:]:
      print("Executing: %s" % arg)
      if arg in globals():
        globals()[arg]()  # assume function and execute
      else:
        eval(arg)  # assume Python code and execute




from __future__ import print_function

from math import sqrt
import numpy
from theano import tensor as T
import theano
from returnn.log import log
from returnn.theano.util import time_batch_make_flat, tiled_eye
from returnn.util.basic import as_str
import json


class Container(object):
  rng_seed = 1234
  layer_class = None

  @classmethod
  def initialize_rng(cls):
    cls.rng = numpy.random.RandomState(cls.rng_seed)

  def __init__(self, layer_class=None, name="", network=None,
               train_flag=False, eval_flag=False, depth=1, consensus="flat",
               forward_weights_init=None, bias_init=None, weight_clip=0.0, cost=None,
               recurrent_weights_init=None,
               substitute_param_expr=None):
    """
    :param str layer_class: name of layer type, e.g. "hidden", "recurrent", "lstm" or so. see LayerClasses.
    :param str name: custom layer name, e.g. "hidden_2"
    :param Network.LayerNetwork network: the network which we will be part of
    :param str forward_weights_init: see self.create_forward_weights()
    :param str bias_init: see self.create_bias()
    """
    self.params = {}; """ :type: dict[str,theano.compile.sharedvalue.SharedVariable] """
    self.attrs = {}; """ :type: dict[str,str|float|int|bool|dict] """
    self.device = None
    if layer_class:
      self.layer_class = as_str(layer_class.encode("utf8"))
    self.name = as_str(name.encode("utf8"))
    self.train_flag = train_flag
    self.eval_flag = eval_flag
    self.depth = depth
    if depth != 1:
      self.set_attr('depth', depth)
    if consensus != "flat":
      self.set_attr('consensus', consensus)
    self.network = network
    if forward_weights_init:
      self.set_attr("forward_weights_init", forward_weights_init)
    self.forward_weights_init = forward_weights_init or "random_normal()"
    if recurrent_weights_init:
      self.set_attr("recurrent_weights_init", recurrent_weights_init)
    self.recurrent_weights_init = recurrent_weights_init or "random_uniform()"
    if bias_init:
      self.set_attr("bias_init", bias_init)
    self.bias_init = bias_init or "zeros()"
    if substitute_param_expr:
      self.set_attr("substitute_param_expr", substitute_param_expr)
    self.substitute_param_expr = substitute_param_expr
    if weight_clip:
      self.set_attr('weight_clip', weight_clip)
    if cost:
      self.set_attr('cost', cost)

  def __repr__(self):
    return "<%s class:%s name:%s>" % (self.__class__, self.layer_class, self.name)

  def dot(self, vec, mat):
    if self.depth == 1:
      return T.dot(vec, mat)
    else:
      return T.tensordot(vec, mat, 1)

  def save(self, head):
    """
    :type head: h5py.File
    """
    grp = head.create_group(self.name)
    grp.attrs['class'] = self.layer_class
    for p in self.params.keys():
      value = self.params[p].get_value()
      dset = grp.create_dataset(p, value.shape, dtype='f')
      dset[...] = value
    for p, v in self.attrs.items():
      if isinstance(v, (dict, list, tuple)):
        v = json.dumps(v, sort_keys=True)
      try:
        grp.attrs[p] = v
      except TypeError:
        print("warning: invalid type of attribute %r (%s) in layer %s" % (p, type(v), self.name), file=log.v3)

  def load(self, head):
    """
    :type head: h5py.File
    """
    try:
      grp = head[self.name]
    except Exception:
      print("warning: unable to load parameters for layer", self.name, file=log.v3)
      return

    grp_class = as_str(grp.attrs['class'])
    if grp_class == "<unknown_softmax>":
      grp_class = "softmax"  # bug in some RETURNN version. can be ignored.
    if grp_class != self.layer_class:
      from .basic import get_layer_class
      if not get_layer_class(grp_class, raise_exception=False) is get_layer_class(self.layer_class):
        print("warning: invalid layer class (expected " + self.layer_class + " got " + grp.attrs['class'] + ")", file=log.v3)
    for p in self.params:
      if p not in grp:
        print("unable to load parameter %s in %s" % (p, self.name), file=log.v4)
    for p in grp:
      if p in self.params:
        if self.params[p].get_value(borrow=True, return_internal_type=True).shape == grp[p].shape:
          array = grp[p][...]
          assert not (numpy.isinf(array).any() or numpy.isnan(array).any())
          self.params[p].set_value(array)
        else:
          print("warning: invalid layer parameter shape for parameter " + p + " of layer " + self.name + \
            " (expected  " + str(self.params[p].get_value(borrow=True, return_internal_type=True).shape) + \
            " got " + str(grp[p].shape) + ")", file=log.v2)
          #assert self.params[p].get_value(borrow=True, return_internal_type=True).shape == grp[p].shape, \
          #  "invalid layer parameter shape for parameter " + p + " of layer " + self.name + \
          #  " (expected  " + str(self.params[p].get_value(borrow=True, return_internal_type=True).shape) + \
          #  " got " + str(grp[p].shape) + ")"
      else:
        print("unable to match parameter %s in %s" % (p, self.name), file=log.v4)
    #for p in self.attrs.keys():
    #  att = grp.attrs.get(p, None)
    #  if att != None:
    #    self.attrs[p] = att

  def num_params(self):
    return sum([numpy.prod(v.get_value(borrow=True, return_internal_type=True).shape[0:]) for v in self.params.values()])

  def get_params_dict(self):
    """
    :rtype: dict[str,numpy.ndarray|theano.sandbox.cuda.CudaNdArray]
    """
    return {p: v.get_value(borrow=True, return_internal_type=True) for (p, v) in self.params.items()}

  def set_params_by_dict(self, params):
    """
    :type params: dict[str,numpy.ndarray|theano.sandbox.cuda.CudaNdArray]
    """
    for p, v in params.items():
      self_param_shape = self.params[p].get_value(borrow=True, return_internal_type=True).shape
      assert self_param_shape == v.shape, "In %s, param %s shape does not match. Expected %s, got %s." % \
                                          (self, p, self_param_shape, v.shape)
      self.params[p].set_value(v, borrow=True)

  def get_params_vars(self):
    """
    :returns list of shared vars in a well-defined order
    """
    res = []
    for (k, v) in sorted(self.params.items()):
      v.layer = self
      res.append(v)
    return res

  def add_param(self, param, name=""):
    """
    :type param: theano.SharedVariable
    :type name: str
    :rtype: theano.SharedVariable
    """
    if not name:
      name = getattr(param, "name", None)
    if not name:
      name = "param_%d" % len(self.params)
    if self.network:
      substitute = self.network.get_layer_param(layer_name=self.name, param_name=name, param=param)
      if substitute:
        return substitute
    if self.substitute_param_expr:
      substitute = eval(self.substitute_param_expr, {"self": self, "name": name, "value": param})
      if substitute:
        return substitute
    self.params[name] = param
    return param

  def set_attr(self, name, value):
    """
    :param str name: key name
    :param bool|int|float|str|list|dict value: value
    This will be stored in to_json() and save() (in HDF).
    More complex types like list or dict will be encoded as a JSON-str when saved to HDF.
    """
    self.attrs[name] = value

  def shared(self, value, name, borrow=True):
    if self.device is None:
      return theano.shared(value=value, borrow=borrow, name=name)
    return theano.shared(value=value, borrow=borrow, name=name, target=self.device)

  def create_bias(self, n, prefix='b', name="", init_eval_str=None):
    """
    :param int n: output dimension
    :rtype: theano.shared
    """
    if not name:
      name = "%s_%s" % (prefix, self.name)
      if name in self.params:
        name += "_%i" % len(self.params)
    if not init_eval_str:
      init_eval_str = self.bias_init
    if self.depth > 1:
      size = (self.depth, n)
    else:
      size = (n,)
    def random_normal(scale, loc=0.0):
      return self.rng.normal(loc=loc, scale=scale, size=size)
    def random_uniform(l, loc=0.0):
      return self.rng.uniform(low=-l + loc, high=l + loc, size=size)
    import returnn.config as config_mod
    import returnn.util.basic as util
    try:
      config = config_mod.get_global_config()
    except Exception:
      config = None
    else:
      config = util.DictAsObj(config.typed_dict)
    eval_locals = {
      "numpy": numpy,
      "rng": self.rng,
      "config": config,
      "self": self,
      "n": n,
      "name": name,
      "sqrt": numpy.sqrt,
      "log": numpy.log,
      "zeros": (lambda: numpy.zeros(size, dtype=theano.config.floatX)),
      "random_normal": random_normal,
      "random_uniform": random_uniform
    }
    values = eval(init_eval_str, eval_locals)
    values = numpy.asarray(values, dtype=theano.config.floatX)
    assert values.shape == (n,)
    return self.shared(values, name)

  def create_random_normal_weights(self, n, m, scale=None, name=None):
    if name is None: name = self.name
    if not scale:
      scale =  numpy.sqrt((n + m) / 12.)
    else:
      scale = numpy.sqrt(scale / 12.)
    if self.depth > 1:
      values = numpy.asarray(self.rng.normal(loc=0.0, scale=1.0 / scale, size=(n, self.depth, m)), dtype=theano.config.floatX)
    else:
      values = numpy.asarray(self.rng.normal(loc=0.0, scale=1.0 / scale, size=(n, m)), dtype=theano.config.floatX)
    return self.shared(values, name)

  def create_random_uniform_weights(self, n, m, p=None, p_add=None, l=None, name=None, depth=None):
    if not depth: depth = self.depth
    if name is None: name = 'W_' + self.name
    assert not (p and l)
    if not p: p = n + m
    if p_add: p += p_add
    if not l: l = sqrt(6.) / sqrt(p)  # 1 / sqrt(p)
    if depth > 1:
      values = numpy.asarray(self.rng.uniform(low=-l, high=l, size=(n, depth, m)), dtype=theano.config.floatX)
    else:
      values = numpy.asarray(self.rng.uniform(low=-l, high=l, size=(n, m)), dtype=theano.config.floatX)
    return self.shared(values, name)

  def create_random_uniform_weights1(self, n, m, p=None, l=None, name=None):
    if name is None: name = 'W_' + self.name
    assert not (p and l)
    if not p: p = n + m
    if not l: l = sqrt(6.) / sqrt(p)  # 1 / sqrt(p)
    values = numpy.asarray(self.rng.uniform(low=-l, high=l, size=(n, m)), dtype=theano.config.floatX)
    return self.shared(values, name)

  def create_random_uniform_weights2(self, n, m=None, name=None):
    if name is None: name = 'W_' + self.name
    l = sqrt(1. / n)
    shape = [n]
    if m: shape += [m]
    values = numpy.asarray(self.rng.uniform(low=-l, high=l, size=shape), dtype=theano.config.floatX)
    return self.shared(values, name)

  def create_random_unitary_weights(self, n, m, name=None):
    x = self.rng.randn(n, m)
    u, s, v = numpy.linalg.svd(x, full_matrices=0)
    if u.shape == (n, m):
      x = u
    else:
      x = v
    assert x.shape == (n, m)
    x = x.astype(theano.config.floatX)
    return self.shared(x, name)

  def create_random_unitary_tiled_weights(self, n, m, name=None):
    if n > m:
      transpose = True
      n, m = m, n  # n < m
    else:  # n <= m
      transpose = False
    fac = ((m - 1) // n) + 1
    def make_tile():
      x = self.rng.randn(n, n)
      u, s, v = numpy.linalg.svd(x)
      assert u.shape == (n, n)
      return u
    x = numpy.concatenate([make_tile() for i in range(fac)], axis=1)
    assert x.shape == (n, fac * n)
    x = x[:, :m]
    assert x.shape == (n, m)
    if transpose:
      x = x.T
    x = x.astype(theano.config.floatX)
    return self.shared(x, name)

  def _create_eval_weights(self, n, m, name, default_name_prefix, init_eval_str):
    """
    :param int n: input dimension
    :param int m: output dimension
    :param str|None name: layer name
    :rtype: theano.shared
    """
    if not name: name = "%s_%s_%i" % (default_name_prefix, self.name, len(self.params))
    from returnn.config import get_global_config
    import returnn.util.basic as util
    try:
      config = get_global_config()
    except Exception:
      config = None
    else:
      config = util.DictAsObj(config.typed_dict)
    eval_locals = {
      "numpy": numpy,
      "theano": theano,
      "rng": self.rng,
      "config": config,
      "self": self,
      "n": n,
      "m": m,
      "name": name,
      "sqrt": numpy.sqrt,
      "eye": (lambda N=n, M=m: numpy.eye(N, M, dtype=theano.config.floatX)),
      "random_normal": (
      lambda scale=None, **kwargs: self.create_random_normal_weights(n, m, scale=scale, name=name, **kwargs)),
      "random_uniform": (
      lambda l=None, p=None, **kwargs: self.create_random_uniform_weights(n, m, p=p, l=l, name=name, **kwargs)),
      "random_unitary": (lambda **kwargs: self.create_random_unitary_weights(n, m, name=name, **kwargs)),
      "random_unitary_tiled": (lambda **kwargs: self.create_random_unitary_tiled_weights(n, m, name=name, **kwargs))
    }
    v = eval(init_eval_str, eval_locals)
    if isinstance(v, numpy.ndarray):
      v = numpy.asarray(v, dtype=theano.config.floatX)
      v = self.shared(v, name)
    assert isinstance(v, theano.compile.SharedVariable)
    assert v.ndim == 2
    vshape = v.get_value(borrow=True, return_internal_type=True).shape
    assert vshape == (n, m)
    return v

  def _create_eval_params(self, shape, name, init_eval_str=None):
    assert 1 <= len(shape) <= 2
    if len(shape) == 1:
      return self.create_bias(n=shape[0], init_eval_str=init_eval_str, name=name)
    else:
      return self._create_eval_weights(
        n=shape[0], m=shape[1],
        init_eval_str=init_eval_str or "random_uniform()",
        default_name_prefix="W", name=name)

  def create_forward_weights(self, n, m, name=None):
    """
    :param int n: input dimension
    :param int m: output dimension
    :param str|None name: layer name
    :rtype: theano.shared
    """
    return self._create_eval_weights(n=n, m=m, name=name, default_name_prefix="W", init_eval_str=self.forward_weights_init)

  def create_recurrent_weights(self, n, m, name=None):
    """
    :param int n: input dimension
    :param int m: output dimension
    :param str|None name: layer name
    :rtype: theano.shared
    """
    return self._create_eval_weights(n=n, m=m, name=name, default_name_prefix="W_re", init_eval_str=self.recurrent_weights_init)

  @classmethod
  def guess_source_layer_name(cls, layer_name):
    # Any model created via NetworkDescription has SourceLayer with empty name as a source.
    # Guess the real source layer name from our name, if it matches the scheme, e.g. "hidden_N_fw".
    import re
    m = re.search("^.*?([0-9]+)[^0-9]*$", layer_name)
    if m:
      nr = int(m.group(1))
      if nr > 0:
        return "%s%i%s" % (layer_name[:m.start(1)], nr - 1, layer_name[m.end(1):])
    return None

  def to_json(self):
    attrs = self.attrs.copy()
    for k in attrs.keys():
      if isinstance(attrs[k], numpy.bool_):
        attrs[k] = True if attrs[k] else False
      if isinstance(attrs[k], bytes):
        attrs[k] = attrs[k].decode("utf8")
    if 'from' in attrs:
      if attrs['from'] == 'data':
        attrs.pop('from', None)
      elif attrs['from'] == '':
        guessed = self.guess_source_layer_name(self.name)
        if guessed:
          attrs['from'] = [guessed]
        else:
          attrs.pop('from', None)
      else:
        attrs['from'] = attrs['from'].split(',')
    return attrs


class SourceLayer(Container):
  layer_class = "source"
  recurrent = False

  def __init__(self, n_out, x_out=None, delay=0, sparse=False, name="", network=None, eval_flag=False,
               data_key=None,  # if we don't want to use "data" but something else. via y_in
               # These will be given if we initialize via JSON.
               sources=None, dropout=0, train_flag=None, mask=None, index=None, y_in=None, dtype=None):
    super(SourceLayer, self).__init__(layer_class=self.layer_class, name=name)
    if data_key is not None:
      assert x_out is None
      assert network
      assert dtype
      network.use_target(target=data_key, dtype=dtype)
      x_out = network.y[data_key]
      n_out = network.n_out[data_key][0]
      index = network.j[data_key]
    if x_out is None:
      assert network is not None
      x_out = network.x
    assert not sources, 'specify `"from": "null"` in json'  # or just ignore?
    assert dropout == 0
    if getattr(x_out.tag, "test_value", None) is None:
      if not sparse:
        x_out.tag.test_value = numpy.random.rand(3,2,n_out).astype('float32')
    if index and getattr(index.tag, "test_value", None) is None:
      index.tag.test_value = numpy.ones((3,2), dtype='int8')
    if not delay:
      self.output = x_out
    else:
      self.output = T.inc_subtensor(T.zeros_like(x_out)[delay:], x_out[:-delay])
    self.set_attr('n_out', n_out)
    self.set_attr('sparse', sparse)
    self.set_attr('delay', delay)
    self.index = index
    self.device = 'cpu'
    self.eval_flag = eval_flag

  def make_constraints(self):
    return 0

  def cost(self):
    return None, None

  def errors(self):
    """
    :rtype: theano.Variable
    """
    return None

  def transfer_output(self, device):
    pass


class Layer(Container):
  recurrent = False

  def __init__(self, sources, n_out, index, y_in=None, target=None, target_index=None,
               sparse=False, cost_scale=1.0, input_scale=1.0,
               L1=0.0, L2=0.0, L2_eye=None, varreg=0.0,
               output_L2_reg=0.0, output_entropy_reg=0.0, output_entropy_exp_reg=0.0,
               with_bias=True,
               mask="unity", dropout=0.0, batch_drop=False, batch_norm=False, bn_use_sample=False, layer_drop=0.0, residual=False,
               carry=False,
               sparse_filtering=False, gradient_scale=1.0, trainable=True, device=None,
               dtype='float32',
               **kwargs):
    """
    :param list[NetworkBaseLayer.Layer] sources: list of source layers
    :param int n_out: output dim of W_in and dim of bias
    :param float L1: l1-param-norm regularization
    :param float L2: l2-param-norm regularization
    :param str mask: "unity" or "dropout"
    :type dropout: float
    """
    super(Layer, self).__init__(**kwargs)
    self.index = index
    self.sources = sources; ":type: list[Layer]"
    self.num_sources = len(sources)
    self.D = max([s.D for s in sources if isinstance(s,Layer)] + [0])
    if mask is None: mask = 'none'
    self.set_attr('mask', mask)
    self.set_attr('dropout', dropout)
    self.set_attr('sparse', sparse)
    self.set_attr('bn_use_sample', bn_use_sample)
    self.set_attr('sparse_filtering', sparse_filtering)
    if not trainable:
      self.set_attr('trainable', trainable)  # only store if not default
      self.gradient_scale = 0.0  # just to be sure
    else:
      self.gradient_scale = gradient_scale
    if gradient_scale != 1.0:
      self.set_attr('gradient_scale', gradient_scale)
    self.set_attr('layer_drop', layer_drop)
    assert not carry, "not supported anymore"
    self.set_attr('residual', residual)
    self.set_attr('n_out', n_out)
    self.set_attr('L1', L1)
    self.set_attr('L2', L2)
    if L2_eye:
      self.set_attr('L2_eye', L2_eye)
    self.device = device # if device else str(theano.config.device)
    for s in self.sources:
      s.transfer_output(self.device)
    self.set_attr('varreg', varreg)
    if output_L2_reg:
      self.set_attr('output_L2_reg', output_L2_reg)
    if output_entropy_reg:
      self.set_attr('output_entropy_reg', output_entropy_reg)
    if output_entropy_exp_reg:
      self.set_attr('output_entropy_exp_reg', output_entropy_exp_reg)
    self.set_attr('batch_norm', batch_norm)
    self.set_attr('input_scale', input_scale)
    if y_in is not None:
      self.y_in = {}
      for k in y_in:
        if not isinstance(y_in[k], T.Variable): continue
        self.y_in[k] = time_batch_make_flat(y_in[k])  # TODO: better not flatten here...
        self.y_in[k].n_out = getattr(y_in[k], "n_out", None)
    else:
      self.y_in = None
    self.constraints = T.constant(0)
    if target:
      self.set_attr('target', target)
    if target_index:
      self.set_attr('target_index', target_index)
      assert target_index in self.network.j
      self.index = index = self.network.j[target_index]
    if cost_scale != 1:
      self.set_attr("cost_scale", cost_scale)
    if with_bias:
      self.b = self.add_param(self.create_bias(n_out), 'b_%s'%self.name)
    else:
      self.set_attr('with_bias', False)
      self.b = numpy.float32(0)
    self.mass = T.constant(1., name = "mass_%s" % self.name, dtype='float32')
    self.masks = [None] * len(self.sources)
    assert mask in ['dropout', 'unity', 'none'], "invalid mask: %s" % mask
    if mask == "dropout" or (mask == 'none' and dropout > 0):
      assert 0.0 < dropout < 1.0
      # If we apply this mass during training then we don't need any mask or mass for testing.
      # The expected weight should be 1 in
      #   E[x] = mass * (1-dropout)
      # so mass has to be 1 / (1 - dropout).
      self.mass = T.constant(1.0 / (1.0 - dropout), dtype='float32')
      from theano.sandbox.rng_mrg import MRG_RandomStreams as RandomStreams
      srng = RandomStreams(self.rng.randint(1234) + 1)
      if self.depth > 1:
        self.masks = [T.cast(srng.binomial(n=1, p=1 - dropout, size=(s.attrs['n_out'],self.depth)), theano.config.floatX) for s in self.sources]
      else:
        if batch_drop:
          self.masks = [T.cast(srng.binomial(n=1, p=1 - dropout, size=s.output.shape), theano.config.floatX) for s in self.sources]
        else:
          self.masks = [T.cast(srng.binomial(n=1, p=1 - dropout, size=(s.attrs['n_out'],)), theano.config.floatX) for s in self.sources]
      #this actually looked like dropconnect applied to the recurrent part, but I want to try dropout for the inputs
      #self.mask = T.cast(srng.binomial(n=1, p=1-dropout, size=(self.attrs['n_out'], self.attrs['n_out'])), theano.config.floatX)

  def concat_units(self, other, axis = 1):
    assert other.layer_class == self.layer_class, "unable to concatenate %s (%s) to %s (%s)" % (other.name, other.layer_class, self.name, self.layer_class)
    for p in other.params.keys():
      if p != 'b':
        self.params[p].set_value(numpy.concatenate((self.params[p].get_value(), other.params[p].get_value()), axis = min(len(self.params[p].get_value().shape) - 1, axis)))
    if axis == 1: self.set_attr('n_out', self.attrs['n_out'] + other.arrs['n_out'])

  def output_index(self):
    from theano.ifelse import ifelse
    index = self.index
    if self.sources:
      # In some cases, e.g. forwarding, the target index (for "classes") might have shape[0]==0.
      # Or shape[0]==1 with index[0]==0. See Dataset.shapes_for_batches().
      # Use source index in that case.
      have_zero = ifelse(T.lt(index.shape[0], 1), 1, T.cast(T.le(index.shape[0], 1) * T.eq(T.sum(index[0]), 0), 'int8'))
      index = ifelse(have_zero, self.sources[0].index, index)
    return index

  def find_data_layer(self):
    for l in self.sources:
      if isinstance(l, SourceLayer):
        return l
      if isinstance(l, Layer):
        s = l.find_data_layer()
        if s is not None:
          return s
    return None

  def add_param(self, param, name="", constraints=True,
                custom_update=None, custom_update_normalized=False, custom_update_exp_average=0,
                custom_update_condition=None, custom_update_accumulate_batches=None, live_update=None):
    """
    :type param: theano.SharedVariable
    :type name: str
    :rtype: theano.SharedVariable
    """
    param = super(Layer, self).add_param(param, name)
    param.live_update = live_update
    if custom_update:
      # Handled in Device and Updater.
      param.custom_update = custom_update
      param.custom_update_normalized = custom_update_normalized
      param.custom_update_exp_average = custom_update_exp_average
      param.custom_update_condition = custom_update_condition
      param.custom_update_accumulate_batches = custom_update_accumulate_batches
    if constraints:
      if 'L1' in self.attrs and self.attrs['L1'] > 0:
        self.constraints += T.constant(self.attrs['L1'], name="L1", dtype='floatX') * abs(param).sum()
      if 'L2' in self.attrs and self.attrs['L2'] > 0:
        self.constraints += T.constant(self.attrs['L2'], name="L2", dtype='floatX') * (param**2).sum()
      if self.attrs.get('L2_eye', 0) > 0:
        L2_eye = T.constant(self.attrs['L2_eye'], name="L2_eye", dtype='floatX')
        if param.ndim == 2:
          eye = tiled_eye(param.shape[0], param.shape[1], dtype=param.dtype)
          self.constraints += L2_eye * ((param - eye)**2).sum()
        else:  # standard L2
          self.constraints += L2_eye * (param**2).sum()
      if 'varreg' in self.attrs and self.attrs['varreg'] > 0:
        self.constraints += self.attrs['varreg'] * (1.0 * T.sqrt(T.var(param)) - 1.0 / numpy.sum(param.get_value().shape))**2
    return param

  def get_branching(self):
    return sum([W.get_value().shape[0] for W in self.W_in]) + 1

  def get_energy(self):
    energy =  self.b / self.attrs['n_out']
    for W in self.W_in:
      energy += T.sum(W, axis = 0)
    return energy

  def make_constraints(self):
    c = self.constraints
    f32_index = T.cast(self.index, "float32")
    if self.attrs.get('output_L2_reg', 0.0):
      # Note: For the output layer, it might even make sense to use a negative factor. http://www.danielpovey.com/files/2016_interspeech_mmi.pdf
      assert self.output.ndim == 3
      l2 = f32_index * T.sum(T.sqr(self.output), axis=2)
      c += numpy.float32(self.attrs.get('output_L2_reg', 0.0)) * T.sum(l2)
    if self.attrs.get('output_entropy_reg', 0.0):
      assert self.output.ndim == 3
      epsilon = numpy.float32(1e-10)
      entropy = f32_index * (-T.sum(self.output * T.log(self.output + epsilon), axis=2))
      c += numpy.float32(self.attrs.get('output_entropy_reg', 0.0)) * T.sum(entropy)
    if self.attrs.get('output_entropy_exp_reg', 0.0):
      assert self.output.ndim == 3
      entropy = f32_index * (-T.sum(T.exp(self.output) * self.output, axis=2))
      c += numpy.float32(self.attrs.get('output_entropy_exp_reg', 0.0)) * T.sum(entropy)
    return c

  def make_consensus(self, networks, axis=2):
    cns = self.attrs['consensus']
    if cns == 'max':
      return T.max(networks, axis=axis)
    elif cns == 'min':
      return T.min(networks, axis=axis)
    elif cns == 'mean':
      return T.mean(networks, axis=axis)
    elif cns == 'flat':
      if self.depth == 1:
        return networks
      if axis == 2:
        return networks.flatten(ndim=3)
        #return T.reshape(networks, (networks.shape[0], networks.shape[1], T.prod(networks.shape[2:]) ))
      else:
        return networks.flatten(ndim=2) # T.reshape(networks, (networks.shape[0], T.prod(networks.shape[1:]) ))
    elif cns == 'sum':
      return T.sum(networks, axis=axis, acc_dtype=theano.config.floatX)
    elif cns == 'prod':
      return T.prod(networks, axis=axis)
    elif cns == 'var':
      return T.var(networks, axis=axis)
    elif cns == 'project':
      p = self.add_param(self.create_random_uniform_weights(self.attrs['n_out'], 1, self.attrs['n_out'] + self.depth + 1))
      return T.tensordot(p, networks, [[1], [axis]])
    elif cns == 'random':
      idx = self.rng.random_integers(size=(1,), low=0, high=self.depth)
      if axis == 0: return networks[idx]
      if axis == 1: return networks[:,idx]
      if axis == 2: return networks[:,:,idx]
      if axis == 3: return networks[:,:,:,idx]
      assert False, "axis too large"
    else:
      assert False, "consensus method unknown: " + cns

  def batch_norm(self, h, dim, use_shift=True, use_std=True, use_sample=0.0, force_sample=False, index=None,
                 sample_mean=None, gamma=None, beta=None, depth_norm=False):
    x = h
    if h.ndim == 3:
      if index is None: index = self.index
      x = h.reshape((h.shape[0]*h.shape[1],h.shape[2]))[(index.flatten()>0).nonzero()]
    elif h.ndim == 4: # index is sizes here
      assert index is not None
      x = h.reshape((h.shape[0] * h.shape[1] * h.shape[2], h.shape[3]))
      #x = x[(T.gt(x,numpy.float32(0))>0).nonzero()]
    mean = T.mean(x,axis=0)
    std = T.sqrt(T.mean((x - mean)**2,axis=0))
    if sample_mean is None:
      sample_mean = self.add_param(theano.shared(numpy.zeros((dim,), 'float32'), '%s_%s_mean' % (self.name,h.name)),
                                   custom_update=mean,
                                   custom_update_normalized=True)
    self.sample_mean = sample_mean
    sample_std = T.sqrt(T.mean((x - sample_mean)**2,axis=0))
    if not self.train_flag and not force_sample:
      use_sample = 1.0
    mean = T.constant(1.-use_sample,'float32') * mean + T.constant(use_sample,'float32') * sample_mean
    std = T.constant(1.-use_sample,'float32') * std + T.constant(use_sample,'float32') * sample_std
    if h.ndim == 3:
      mean = mean.dimshuffle('x','x',0).repeat(h.shape[0],axis=0).repeat(h.shape[1],axis=1)
      std = std.dimshuffle('x', 'x', 0).repeat(h.shape[0],axis=0).repeat(h.shape[1],axis=1)
    elif h.ndim == 4:
      mean = mean.dimshuffle('x', 'x', 'x', 0).repeat(h.shape[0], axis=0).repeat(h.shape[1], axis=1).repeat(h.shape[2], axis=2)
      std = std.dimshuffle('x', 'x', 'x', 0).repeat(h.shape[0], axis=0).repeat(h.shape[1], axis=1).repeat(h.shape[2], axis=2)
    else:
      mean = mean.dimshuffle('x', 0).repeat(h.shape[0], axis=0)
      std = std.dimshuffle('x', 0).repeat(h.shape[0], axis=0)
    bn = (h - mean) / (std + numpy.float32(1e-10))
    if use_std:
      if gamma is None:
        gamma = self.add_param(self.shared(numpy.zeros((dim,), 'float32') + numpy.float32(0.1), "%s_%s_gamma" % (self.name,h.name)))
      self.gamma = gamma
      if h.ndim == 3:
        bn *= gamma.dimshuffle('x','x',0).repeat(h.shape[0],axis=0).repeat(h.shape[1],axis=1)
      elif h.ndim == 4:
        bn *= gamma.dimshuffle('x', 'x', 'x', 0).repeat(h.shape[0], axis=0).repeat(h.shape[1], axis=1).repeat(h.shape[2], axis=2)
      else:
        bn *= gamma.dimshuffle('x', 0).repeat(h.shape[0], axis=0)
    if use_shift:
      if beta is None:
        beta = self.add_param(self.shared(numpy.zeros((dim,), 'float32'), "%s_%s_beta" % (self.name,h.name)))
      self.beta = beta
      bn += beta
    if depth_norm:
      bn = bn / (T.sqrt(2)**self.D)
    return bn

  def make_output(self, output, collapse = True, sample_mean=None, gamma=None):
    self.output = output
    if collapse and self.depth > 1:
      self.output = self.make_consensus(self.output)
      if self.attrs['consensus'] == 'flat':
        self.attrs['n_out'] *= self.depth
    if self.attrs['batch_norm']:
      self.output = self.batch_norm(self.output, self.attrs['n_out'], sample_mean=sample_mean, gamma=gamma, use_sample=self.attrs['bn_use_sample'])
    if self.attrs['residual']:
      from .hidden import concat_sources
      z, n_in = concat_sources(self.sources, unsparse=True, expect_source=False)
      assert n_in == self.attrs['n_out']
      self.output += z
    if self.attrs['layer_drop'] > 0.0:
      # Stochastic Depth, http://arxiv.org/abs/1603.09382
      from .hidden import concat_sources
      z, n_in = concat_sources(self.sources, unsparse=True, expect_source=False)
      n_out = self.attrs['n_out']
      if n_in != n_out:
        print("Layer drop with additional projection %i -> %i" % (n_in, n_out), file=log.v4)
        if n_in > 0:
          self.W_drop = self.add_param(self.create_forward_weights(n_in, n_out, name="W_drop_%s" % self.name))
          z = T.dot(z, self.W_drop)
        else:
          z = 0
      if self.train_flag:
        from theano.sandbox.rng_mrg import MRG_RandomStreams as RandomStreams
        rng = RandomStreams(self.rng.randint(1234) + 1)
        import theano.ifelse
        drop = rng.binomial(n=1, p=self.attrs['layer_drop'], size=(1,), dtype='int8')[0]
        # drop = theano.printing.Print("drop")(drop)
        self.output = theano.ifelse.ifelse(drop, z, self.output)
      else:
        drop = self.attrs['layer_drop']
        self.output = numpy.float32(drop) * z + numpy.float32(1.0 - drop) * self.output
    if self.attrs['sparse']:
      self.output = T.argmax(self.output, axis=-1, keepdims=True)
    if self.attrs['sparse_filtering']:
      # https://dlacombejr.github.io/programming/2015/09/13/sparse-filtering-implemenation-in-theano.html
      fs = T.sqrt(self.output ** 2 + 1e-8)    # numerical stability
      l2fs = T.sqrt(T.sum(fs ** 2, axis=1))   # l2 norm of row
      nfs = fs / l2fs.dimshuffle(0, 'x')      # normalize rows
      l2fn = T.sqrt(T.sum(nfs ** 2, axis=0))  # l2 norm of column
      self.output = nfs / l2fn.dimshuffle('x', 0)   # normalize columns
    self.output.name = "%s.output" % self.name
    self._output = output

  def transfer_output(self, device):
    if device is None:
      device = str(theano.config.device)
      if self.device is None:
        return self.output
    if device != self.device:
      self.output = self._output.transfer(device) # requires Theano 0.8
    else:
      self.output = self._output

  def to_json(self):
    attrs = super(Layer, self).to_json()
    attrs['class'] = self.layer_class
    return attrs

  def cost(self):
    """
    :rtype: (theano.Variable | None, dict[theano.Variable,theano.Variable] | None)
    :returns: cost, known_grads
    """
    return None, None

  def cost_scale(self):
    """
    :rtype: theano.Variable
    """
    return T.constant(self.attrs.get("cost_scale", 1.0), dtype="float32")

  def errors(self):
    """
    :rtype: theano.Variable
    """
    return None



#!/usr/bin/env python3

"""
Given BPE codes/vocab, create a lexicon (mapping of words to all possible BPE seqs).
"""

from __future__ import print_function

import typing
from argparse import ArgumentParser
from xml.etree import ElementTree


BpeMergeSymbol = "@@"


class PrefixTree:
  """
  Prefix tree / trie.
  This class represents both a single node and the tree.
  """

  def __init__(self, prefix="", root=None):
    """
    :param str prefix:
    :param PrefixTree|None root:
    """
    self.prefix = prefix
    self.arcs = {}  # type: typing.Dict[str,PrefixTree]
    self.finished = False
    self.bpe_finished = False
    self.is_root = not root
    self.root = root

  def add(self, postfix, root=None):
    """
    :param str postfix:
    :param None|PrefixTree root:
    :rtype: PrefixTree
    """
    if not root:
      if self.is_root:
        root = self
      else:
        assert self.root
        root = self.root
    if postfix == BpeMergeSymbol:
      arc = postfix
      postfix_ = ""
    else:
      arc = postfix[:1]
      postfix_ = postfix[1:]
    if arc in self.arcs:
      child = self.arcs[arc]
    else:
      child = PrefixTree(root=root, prefix=self.prefix + arc)
      self.arcs[arc] = child
    if arc == BpeMergeSymbol and not postfix_:
      self.bpe_finished = True
    if postfix_:
      return child.add(postfix_, root=root)
    else:
      child.finished = True
      return child


class Hyp:
  """
  Represents a hypothesis in the search.
  """

  def __init__(self, bpe_sym_history, cur_node):
    """
    :param list[str] bpe_sym_history:
    :param PrefixTree cur_node:
    """
    self.bpe_sym_history = bpe_sym_history
    self.cur_node = cur_node


class Search:
  """
  Covers the search hyps and the search itself.
  """

  def __init__(self, bpe, word, word_pos=0):
    """
    :param PrefixTree bpe:
    :param str word:
    :param int word_pos:
    """
    self.bpe = bpe
    self.word = word
    self.word_pos = word_pos
    self.hyps = [Hyp(bpe_sym_history=[], cur_node=bpe)]  # type: typing.List[Hyp]
    self.final_bpe_seqs = None  # type: typing.Optional[typing.List[typing.List[str]]]

  def _get_finished(self):
    assert self.word_pos == len(self.word)
    finals = []  # type: typing.List[typing.List[str]]
    for hyp in self.hyps:
      if hyp.cur_node.finished:
        finals.append(hyp.bpe_sym_history + [hyp.cur_node.prefix])
    self.final_bpe_seqs = finals

  def _expand(self):
    assert self.word_pos < len(self.word)
    char = self.word[self.word_pos]
    new_hyps = []  # type: typing.List[Hyp]
    for hyp in self.hyps:
      if hyp.cur_node.bpe_finished:
        next_node = self.bpe.arcs.get(char)
        if next_node:
          new_hyps.append(
            Hyp(
              bpe_sym_history=hyp.bpe_sym_history + [hyp.cur_node.prefix + "@@"],
              cur_node=next_node))
      next_node = hyp.cur_node.arcs.get(char)
      if next_node:
        new_hyps.append(Hyp(bpe_sym_history=hyp.bpe_sym_history, cur_node=next_node))
    self.hyps = new_hyps

  def search(self):
    """
    :return: collection of possible BPE symbol seqs
    :rtype: list[list[str]]
    """
    while self.word_pos < len(self.word):
      self._expand()
      self.word_pos += 1
    self._get_finished()
    return self.final_bpe_seqs


def parse_vocab(filename):
  """
  Can be either pure text file, line-based, or lexicon XML file, or Python vocab dict.

  :param str filename:
  :rtype: list[str]
  """
  if filename.endswith(".gz"):
    import gzip
    raw = gzip.open(filename, "r").read().decode("utf8")
  else:
    raw = open(filename, "r").read()
  if raw.startswith("{"):  # Python dict (str sym -> int idx)
    py_vocab = eval(raw)
    assert isinstance(py_vocab, dict)  # sym -> idx
    labels = {idx: label for (label, idx) in sorted(py_vocab.items())}
    min_label, max_label, num_labels = min(labels), max(labels), len(labels)
    assert 0 == min_label
    if num_labels - 1 < max_label:
      print("Vocab error: not all indices used? max label: %i" % max_label)
      print("unused labels: %r" % ([i for i in range(max_label + 1) if i not in labels],))
    assert num_labels - 1 == max_label
    zero_sym = labels[0]
    assert isinstance(zero_sym, str)
    return [label for (idx, label) in sorted(labels.items())]
  if raw.startswith("<?xml"):  # lexicon XML
    labels = []
    from io import StringIO
    raw_stream = StringIO(raw)
    context = iter(ElementTree.iterparse(raw_stream, events=('start', 'end')))
    _, root = next(context)  # get root element
    for event, elem in context:
      if event == "end" and elem.tag == "lemma":
        for orth_elem in elem.findall("orth"):
          orth = (orth_elem.text or "").strip()
          labels.append(orth)
        root.clear()  # free memory
    return labels
  # Assume line-based. No idea how to to a good sanity check...
  return raw.splitlines()


def xml_prettify(element, indent='  '):
  """
  https://stackoverflow.com/a/38574067/133374 (deleted StackOverflow answer)

  :param ElementTree.Element element:
  :param str indent:
  """
  queue = [(0, element)]  # (level, element)
  while queue:
    level, element = queue.pop(0)
    children = [(level + 1, child) for child in list(element)]
    if children:
      element.text = '\n' + indent * (level + 1)  # for child open
    if queue:
      element.tail = '\n' + indent * queue[0][0]  # for sibling open
    else:
      element.tail = '\n' + indent * (level - 1)  # for parent close
    queue[0:0] = children  # prepend so children come before siblings


def main():
  """
  Main entry.
  """
  arg_parser = ArgumentParser()
  arg_parser.add_argument("--bpe_vocab", required=True)
  arg_parser.add_argument("--word_vocab", required=True)
  arg_parser.add_argument("--unk")
  arg_parser.add_argument("--skip_special", action="store_true")
  arg_parser.add_argument("--lower_case", action="store_true")
  arg_parser.add_argument("--output")
  args = arg_parser.parse_args()

  bpe_syms = parse_vocab(args.bpe_vocab)
  words = parse_vocab(args.word_vocab)
  print("BPE symbols: num %i, first %r" % (len(bpe_syms), bpe_syms[0]))
  print("Words: num %i, first %r" % (len(words), words[0]))

  print("Build BPE prefix tree...")
  bpe = PrefixTree()
  for bpe_sym in bpe_syms:
    bpe.add(bpe_sym)

  print("Build lexicon...")
  xml = ElementTree.Element("lexicon")
  xml_phone_inventory = ElementTree.SubElement(xml, "phoneme-inventory")
  for bpe_sym in bpe_syms:  # each BPE symbol will be a phoneme in the XML
    xml_phone = ElementTree.SubElement(xml_phone_inventory, "phoneme")
    ElementTree.SubElement(xml_phone, "symbol").text = bpe_sym
    ElementTree.SubElement(xml_phone, "variation").text = "context"

  visited_words = set()

  # noinspection PyShadowingNames
  def visit_word(word):
    """
    :param str word:
    """
    if word in visited_words:
      return
    visited_words.add(word)
    bpe_sym_seqs = Search(bpe=bpe, word=word).search()
    if not bpe_sym_seqs:
      print("no BPE seq found for word %r" % word)
      return
    xml_lemma = ElementTree.SubElement(xml, "lemma")
    ElementTree.SubElement(xml_lemma, "orth").text = word
    for bpe_sym_seq in bpe_sym_seqs:
      ElementTree.SubElement(xml_lemma, "phon").text = " ".join(bpe_sym_seq)

  for word in words:
    if args.lower_case:
      word = word.lower()
    if not word:
      continue
    if args.skip_special:
      if word.startswith("[") and word.endswith("]"):
        continue
      if word.startswith("<") and word.endswith(">"):
        continue
    visit_word(word)

  for bpe_sym in bpe_syms:
    if bpe_sym.endswith(BpeMergeSymbol):
      continue
    if bpe_sym not in words:
      continue
    # E.g. special symbols, which were skipped above.
    visit_word(bpe_sym)

  if args.output:
    xml_prettify(xml)
    xml_str = ElementTree.tostring(xml, encoding="utf-8")
    with open(args.output, "wb") as f:
      f.write(xml_str)
    print("Wrote XML:", args.output)
  else:
    print("Specify --output to save the XML.")


if __name__ == '__main__':
  from returnn.util import better_exchook
  better_exchook.install()
  main()




from __future__ import print_function

import _setup_test_env  # noqa
from subprocess import Popen, PIPE, STDOUT, CalledProcessError
import re
import os
import sys
from glob import glob
from nose.tools import assert_less, assert_in
from returnn.util import better_exchook
from returnn.util.basic import which_pip


py = sys.executable
print("Python:", py)


def build_env():
  theano_flags = {key: value for (key, value)
                  in [s.split("=", 1) for s in os.environ.get("THEANO_FLAGS", "").split(",") if s]}
  # First set some sane default for compile dir.
  theano_flags.setdefault("compiledir_format",
                          "compiledir_%(platform)s-%(processor)s-%(python_version)s-%(python_bitwidth)s")
  # Extend compile dir for our tests in a subprocess.
  # This is important because there might be other tests where we import Theano in this
  # process and any subprocess would block because the parent process has locked the compile-dir.
  theano_flags["compiledir_format"] += "--nosetests"
  # Nose-tests will set mode=FAST_COMPILE. We don't want this for our tests as it is way too slow.
  theano_flags["mode"] = "FAST_RUN"
  env_update = os.environ.copy()
  env_update["THEANO_FLAGS"] = ",".join(["%s=%s" % (key, value) for (key, value) in theano_flags.items()])
  return env_update


def run(*args):
  args = list(args)
  print("run:", args)
  # RETURNN by default outputs on stderr, so just merge both together
  p = Popen(args, stdout=PIPE, stderr=STDOUT, env=build_env())
  out, _ = p.communicate()
  if p.returncode != 0:
    print("Return code is %i" % p.returncode)
    print("std out/err:\n---\n%s\n---\n" % out.decode("utf8"))
    raise CalledProcessError(cmd=args, returncode=p.returncode, output=out)
  return out.decode("utf8")


def run_and_parse_last_fer(*args):
  out = run(*args)
  parsed_fer = None
  for l in out.splitlines():
    # example: epoch 5 score: 0.0231807245472 elapsed: 0:00:04 dev: score 0.0137521058997 error 0.00268961807423
    m = re.match("epoch [0-9]+ score: .* dev: .* error ([0-9.]+)\\s?", l)
    if not m:
      continue
    parsed_fer = float(m.group(1))
  assert parsed_fer is not None, "No epoch dev errors found in output: %s\n" % out
  return parsed_fer


def run_config_get_fer(config_filename):
  cleanup_tmp_models(config_filename)
  fer = run_and_parse_last_fer(py, "rnn.py", config_filename, "++log_verbosity", "5")
  cleanup_tmp_models(config_filename)
  return fer


def cleanup_tmp_models(config_filename):
  assert os.path.exists(config_filename)
  from returnn.config import Config
  config = Config()
  config.load_file(config_filename)
  model_filename = config.value('model', '')
  assert model_filename
  # Remove existing models
  assert model_filename.startswith("/tmp/")
  for f in glob(model_filename + ".*"):
    os.remove(f)


class TestDemos(object):

  @classmethod
  def setup_class(cls):
    os.chdir((os.path.dirname(__file__) or ".") + "/..")
    assert os.path.exists("rnn.py")

  def test_demo_task12ax(self):
    fer = run_config_get_fer("demos/demo-theano-task12ax.config")
    assert_less(fer, 0.01)

  def test_demo_iter_dataset_task12ax(self):
    cleanup_tmp_models("demos/demo-theano-task12ax.config")
    out = run(py, "demos/demo-iter-dataset.py", "demos/demo-theano-task12ax.config")
    assert_in("Epoch 5.", out.splitlines())

  def test_demo_returnn_as_framework(self):
    print("Prepare.")
    import subprocess
    import shutil
    from glob import glob
    from returnn.util.basic import get_login_username
    # echo via subprocess, because this stdout as well as the other will always be visible.
    subprocess.check_call(["echo", "travis_fold:start:test_demo_returnn_as_framework"])
    assert os.path.exists("setup.py")
    if glob("dist/*.tar.gz"):
      # we want it unique below
      for fn in glob("dist/*.tar.gz"):
        os.remove(fn)
    if os.path.exists("MANIFEST"):
      os.remove("MANIFEST")  # auto-generated. will be recreated
    if os.path.exists("docs/crnn"):
      os.remove("docs/crnn")  # this is auto-generated, and confuses setup.py sdist
    if os.path.exists("docs/returnn"):
      os.remove("docs/returnn")  # this is auto-generated, and confuses setup.py sdist
    tmp_model_dir = "/tmp/%s/returnn-demo-as-framework" % get_login_username()
    if os.path.exists(tmp_model_dir):
      shutil.rmtree(tmp_model_dir, ignore_errors=True)
    print("setup.py sdist, to create package.")
    subprocess.check_call([py, "setup.py", "sdist"])
    dist_fns = glob("dist/*.tar.gz")
    assert len(dist_fns) == 1
    dist_fn = os.path.abspath(dist_fns[0])
    pip_path = which_pip()
    print("Pip install Returnn.")
    in_virtual_env = hasattr(sys, 'real_prefix')  # https://stackoverflow.com/questions/1871549/
    cmd = [py, pip_path, "install"]
    if not in_virtual_env:
      cmd += ["--user"]
    cmd += ["-v", dist_fn]
    print("$ %s" % " ".join(cmd))
    subprocess.check_call(cmd, cwd="/")
    print("Running demo now.")
    subprocess.check_call([py, "demo-returnn-as-framework.py"], cwd="demos")
    print("Success.")
    subprocess.check_call(["echo", "travis_fold:end:test_demo_returnn_as_framework"])

  def test_demo_sprint_interface(self):
    import subprocess
    # echo via subprocess, because this stdout as well as the other will always be visible.
    subprocess.check_call(["echo", "travis_fold:start:test_demo_sprint_interface"])
    subprocess.check_call([py, os.path.abspath("demos/demo-sprint-interface.py")], cwd="/")
    subprocess.check_call(["echo", "travis_fold:end:test_demo_sprint_interface"])

  def test_returnn_as_framework_TaskSystem(self):
    import subprocess
    # echo via subprocess, because this stdout as well as the other will always be visible.
    subprocess.check_call(["echo", "travis_fold:start:test_returnn_as_framework_TaskSystem"])
    subprocess.check_call([py, os.path.abspath("tests/returnn-as-framework.py"), "test_TaskSystem_Pickler()"], cwd="/")
    subprocess.check_call(["echo", "travis_fold:end:test_returnn_as_framework_TaskSystem"])

  def test_returnn_as_framework_old_style_crnn_TFUtil(self):
    import subprocess
    # echo via subprocess, because this stdout as well as the other will always be visible.
    subprocess.check_call(["echo", "travis_fold:start:test_returnn_as_framework_old_style_crnn_TFUtil"])
    subprocess.check_call([
      py, os.path.abspath("tests/returnn-as-framework.py"),
      "--old-style", "--returnn-package-name", "crnn",
      "test_old_style_import_crnn_TFUtil()"], cwd="/")
    subprocess.check_call(["echo", "travis_fold:end:test_returnn_as_framework_old_style_crnn_TFUtil"])

  def test_returnn_as_framework_old_style_TFUtil(self):
    import subprocess
    # echo via subprocess, because this stdout as well as the other will always be visible.
    subprocess.check_call(["echo", "travis_fold:start:test_returnn_as_framework_old_style_TFUtil"])
    subprocess.check_call([
      py, os.path.abspath("tests/returnn-as-framework.py"), "--old-style", "test_old_style_import_TFUtil()"], cwd="/")
    subprocess.check_call(["echo", "travis_fold:end:test_returnn_as_framework_old_style_TFUtil"])


if __name__ == '__main__':
  better_exchook.install()
  TestDemos.setup_class()
  tests = TestDemos()
  import sys
  fns = sys.argv[1:]
  if not fns:
    fns = [arg for arg in dir(tests) if arg.startswith("test_")]
  for arg in fns:
    f = getattr(tests, arg)
    print("-" * 20)
    print("Run:", f)
    print("-" * 20)
    f()
    print("-" * 20)
    print("Ok.")
    print("-" * 20)
